#ifndef __mVec4_H__
#define __mVec4_H__

/*! 
    \file mVec4.h
    \brief 4D Vector
    \author Douglas H. Cox
*/

//============================================================================
//    REQUIRED HEADERS
//============================================================================
#include "mMath.h"

//============================================================================
//    DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS / CLASS FORWARDS
//============================================================================

//============================================================================
//    FUNCTION PROTOTYPES
//============================================================================

//============================================================================
//    CLASS PROTOTYPES
//============================================================================

//! 4D Vector
/*! mVec4 represents a vector consisting of an array of four float values. */
class MMATH_DLL mVec4
{
public:
  float v[4];   //!< vector data

  //! Default constructor
  /*! No initialization is done here. */
  mVec4() {}

  //! Copy constructor
  mVec4(const mVec4& src)
  { 
    v[0] = src.v[0]; 
    v[1] = src.v[1]; 
    v[2] = src.v[2];
    v[3] = src.v[3];
  }

  //! Constructor from 4 float values
  mVec4(float x, float y, float z, float w)
  {
    v[0] = x; 
    v[1] = y; 
    v[2] = z;
    v[3] = w;
  }

  //! Constructor from an array of 4 float values
  mVec4(float *src)
  {
    v[0] = src[0];
    v[1] = src[1];
    v[2] = src[2];
    v[3] = src[3];
  }

  //! Assignment
  void set(const mVec4& src)
  {
    v[0] = src.v[0]; 
    v[1] = src.v[1]; 
    v[2] = src.v[2];
    v[3] = src.v[3];
  }

  //! Assignment from 4 float values
  void set(float x, float y, float z, float w)
  {
    v[0] = x; 
    v[1] = y; 
    v[2] = z;
    v[3] = w;
  }

  //! Assignment from an arry of 4 float values
  void set(float *src)
  {
    v[0] = src[0];
    v[1] = src[1];
    v[2] = src[2];
    v[3] = src[3];
  }

  //! Assignment
  mVec4& operator = (const mVec4& src)
  { 
    v[0] = src.v[0]; 
    v[1] = src.v[1]; 
    v[2] = src.v[2];
    v[3] = src.v[3];
    return *this;
  }

  //! Accessor
  void get(float& x, float& y, float& z, float& w) const
  {
    x = v[0];
    y = v[1];
    z = v[2];
    w = v[3];
  }

  //! Accessor
  float& operator [] (int i)
  {
    return v[i];
  }

  //! Accessor
  float operator [] (int i) const
  {
    return v[i];
  }

  //! Returns true if this == a
  bool operator == (const mVec4& a) const
  {
    return( v[0] == a.v[0] &&
            v[1] == a.v[1] &&
            v[2] == a.v[2] &&
            v[3] == a.v[3] );
  }

  //! Returns true if this != a
  bool operator != (const mVec4& a) const
  { 
    return !(operator==(a));
  }

  //! Returns true is this is equal to a within some tolerance value.
  bool approxEqual(const mVec4& a, float tol = mZeroTol)
  {
    return( mApproxEqual(v[0], a.v[0], tol) &&
            mApproxEqual(v[1], a.v[1], tol) &&
            mApproxEqual(v[2], a.v[2], tol) &&
            mApproxEqual(v[3], a.v[3], tol) );
  }


  //! Sets each component to the min of this and a.
  void setMin(const mVec4& a)
  {
    if (a.v[0] < v[0]) v[0] = a.v[0];
    if (a.v[1] < v[1]) v[1] = a.v[1];
    if (a.v[2] < v[2]) v[2] = a.v[2];
    if (a.v[3] < v[3]) v[3] = a.v[3];
  }

  //! Sets each component to the max of this and a.
  void setMax(const mVec4& a)
  {
    if (a.v[0] > v[0]) v[0] = a.v[0];
    if (a.v[1] > v[1]) v[1] = a.v[1];
    if (a.v[2] > v[2]) v[2] = a.v[2];
    if (a.v[3] > v[3]) v[3] = a.v[3];
  }

  //! Clamps each component of this to [a, b]
  void clamp(float a, float b)
  {
    v[0] = mClamp(v[0], a, b);
    v[1] = mClamp(v[1], a, b);
    v[2] = mClamp(v[2], a, b);
    v[3] = mClamp(v[3], a, b);
  }

  //! Clamps each component of vec to [a, b]
  void clamp(const mVec4& vec, float a, float b)
  {
    v[0] = mClamp(vec.v[0], a, b);
    v[1] = mClamp(vec.v[1], a, b);
    v[2] = mClamp(vec.v[2], a, b);
    v[3] = mClamp(vec.v[3], a, b);
  }

  //! Addition: this += a
  void add(const mVec4& a)
  {
    v[0] += a.v[0];
    v[1] += a.v[1];
    v[2] += a.v[2];
    v[3] += a.v[3];
  }

  //! Addition: this = a + b
  void add(const mVec4& a, const mVec4& b)
  {
    v[0] = a.v[0] + b.v[0];
    v[1] = a.v[1] + b.v[1];
    v[2] = a.v[2] + b.v[2];
    v[3] = a.v[3] + b.v[3];
  }

  //! Addition: Returns this + a
  mVec4 operator + (const mVec4& a) const
  {
    return mVec4(v[0] + a.v[0],
                 v[1] + a.v[1],
                 v[2] + a.v[2],
                 v[3] + a.v[3]);
  }

  //! Addition: this += a
  mVec4& operator += (const mVec4& a)
  {
    v[0] += a.v[0];
    v[1] += a.v[1];
    v[2] += a.v[2];
    v[3] += a.v[3];
    return *this;
  }

  //! this += a * s
  void addScaled(const mVec4& a, float s)
  {
    v[0] += a.v[0] * s;
    v[1] += a.v[1] * s;
    v[2] += a.v[2] * s;
    v[3] += a.v[3] * s;
  }

  //! this = a + b * s
  void addScaled(const mVec4& a, const mVec4& b, float s)
  {
    v[0] = a.v[0] + b.v[0] * s;
    v[1] = a.v[1] + b.v[1] * s;
    v[2] = a.v[2] + b.v[2] * s;
    v[3] = a.v[3] + b.v[3] * s;
  }

  //! this = a * s1 + b * s2
  void addScaled(const mVec4& a, float s1,
                 const mVec4& b, float s2)
  {
    v[0] = a.v[0] * s1 + b.v[0] * s2;
    v[1] = a.v[1] * s1 + b.v[1] * s2;
    v[2] = a.v[2] * s1 + b.v[2] * s2;
    v[3] = a.v[3] * s1 + b.v[3] * s2;
  }

  //! Subtraction: this -= a
  void sub(const mVec4& a)
  {
    v[0] -= a.v[0];
    v[1] -= a.v[1];
    v[2] -= a.v[2];
    v[3] -= a.v[3];
  }

  //! Subtraction: this = a - b
  void sub(const mVec4& a, const mVec4& b)
  {
    v[0] = a.v[0] - b.v[0];
    v[1] = a.v[1] - b.v[1];
    v[2] = a.v[2] - b.v[2];
    v[3] = a.v[3] - b.v[3];
  }

  //! Subtraction: Returns this - a
  mVec4 operator - (const mVec4& a) const
  {
    return mVec4(v[0] - a.v[0],
                 v[1] - a.v[1],
                 v[2] - a.v[2],
                 v[3] - a.v[3]);
  }

  //! Subtraction: this -= a
  mVec4& operator -= (const mVec4& a)
  {
    v[0] -= a.v[0];
    v[1] -= a.v[1];
    v[2] -= a.v[2];
    v[3] -= a.v[3];
    return *this;
  }

  //! Scales each component by s.
  void scale(float s)
  {
    v[0] *= s; 
    v[1] *= s; 
    v[2] *= s;
    v[3] *= s;
  }

  //! Scales each component of a by s.
  void scale(const mVec4& a, float s)
  {
    v[0] = a.v[0] * s;
    v[1] = a.v[1] * s;
    v[2] = a.v[2] * s;
    v[3] = a.v[3] * s;
  }

  //! Scales each component by s.
  mVec4& operator *= (float s)
  {
    v[0] *= s; 
    v[1] *= s; 
    v[2] *= s;
    v[3] *= s;
    return *this;
  }

  //! Negate: this = -this
  void neg()
  {
    v[0] = -v[0];
    v[1] = -v[1];
    v[2] = -v[2];
    v[3] = -v[3];
  }

  //! Negate: this = -a
  void neg(const mVec4& a)
  {
    v[0] = -a.v[0];
    v[1] = -a.v[1];
    v[2] = -a.v[2];
    v[3] = -a.v[3];
  }

  //! Negate: Returns -this
  mVec4 operator - () const
  {
    return mVec4(-v[0], -v[1], -v[2], -v[3]);
  }
  
  //! Component multiplication
  void compMult(const mVec4& a)
  {
    v[0] *= a[0];
    v[1] *= a[1];
    v[2] *= a[2];
    v[3] *= a[3];
  }

  //! Component multiplication
  void compMult(const mVec4& a, const mVec4& b)
  {
    v[0] = a[0] * b[0];
    v[1] = a[1] * b[1];
    v[2] = a[2] * b[2];
    v[3] = a[3] * b[3];
  }

  //! Component division
  void compDiv(const mVec4& a)
  {
    v[0] /= a[0];
    v[1] /= a[1];
    v[2] /= a[2];
    v[3] /= a[3];
  }

  //! Component division
  void compDiv(const mVec4& a, const mVec4& b)
  {
    v[0] = a[0] * b[0];
    v[1] = a[1] * b[1];
    v[2] = a[2] * b[2];
    v[3] = a[3] * b[3];
  }

  //! Returns the distance squared between this and a.
  float distance2(const mVec4& a) const
  {
    return( mSquare(v[0] - a.v[0]) +
            mSquare(v[1] - a.v[1]) +
            mSquare(v[2] - a.v[2]) +
            mSquare(v[3] - a.v[3]) );
  }

  //! Returns the distance between this and a.
  float distance(const mVec4& a) const
  {
    return( mSqrt(distance2(a)) );
  }

  //! Returns the length squared of this.
  float length2() const
  {
    return( mSquare(v[0]) + 
            mSquare(v[1]) + 
            mSquare(v[2]) +
            mSquare(v[3]) );
  }

  //! Returns the length of this.
  float length() const
  {
    return( mSqrt(length2()) );
  }

  //! Returns the dot product of this and a.
  float dot(const mVec4& a) const
  {
    return( v[0] * a.v[0] +
            v[1] * a.v[1] +
            v[2] * a.v[2] +
            v[3] * a.v[3] );
  }

  //! Normalizes this and returns the previous length.
  float normalize()
  {
    float l(length());
    if (l != 0.0f) scale(1.0f / l);
    return l;
  }

  //! Linearly interpolates between two vectors.
  void lerp(const mVec4& a, const mVec4& b, float t)
  {
    v[0] = a.v[0] + t * (b.v[0] - a.v[0]);
    v[1] = a.v[1] + t * (b.v[1] - a.v[1]);
    v[2] = a.v[2] + t * (b.v[2] - a.v[2]);
    v[3] = a.v[3] + t * (b.v[3] - a.v[3]);
  }

};

// Common constant values.
const mVec4 mZero4(0.0f, 0.0f, 0.0f, 0.0f);
const mVec4 mOne4(1.0f, 1.0f, 1.0f, 1.0f);

//****************************************************************************
//**    END HEADER mVec4.h
//****************************************************************************
#endif // __mVec4_H__
