//****************************************************************************
//**
//**    mOCS.cpp
//**    Implementation for mOCS.
//**    (C) 2000 - Douglas H. Cox
//**
//****************************************************************************
//============================================================================
//    HEADERS
//============================================================================
#include "mOCS.h"

#include "mMatrix.h"


//============================================================================
//    PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
//============================================================================

//============================================================================
//    PRIVATE STRUCTURES / UTILITY CLASSES
//============================================================================

//============================================================================
//    PRIVATE DATA
//============================================================================

//============================================================================
//    PRIVATE FUNCTIONS
//============================================================================

//============================================================================
//    FUNCTIONS
//============================================================================

//============================================================================
//    CLASS METHODS
//============================================================================


//=============================================================
// mOCS::mOCS
mOCS::mOCS(const mOCS& src) : o(src.o), 
                              f(src.f), 
                              s(src.s), 
                              type(src.type)
{
}


//=============================================================
// mOCS::mOCS
mOCS::mOCS(const mVec3& origin, 
           const mFrame& frame, 
           const mVec3& scale,
           unsigned type) 
{
  (type & ORIGIN) ? o.set(origin) : o.set(0.0f, 0.0f, 0.0f);
  (type & FRAME)  ? f.set(frame)  : f.makeIdentity();
  (type & SCALE)  ? s.set(scale)  : s.set(1.0f, 1.0f, 1.0f);

  this->type = type;
}

//=============================================================
// mOCS::makeIdentity
void mOCS::makeIdentity()
{
  if (type == IDENTITY) return;

  o.set(0.0f, 0.0f, 0.0f);
  if (type & FRAME)
    f.makeIdentity();
  s.set(1.0f, 1.0f, 1.0f);
  type = IDENTITY;
}

//=============================================================
// mOCS::makeOrigin
void mOCS::makeOrigin(float x, float y, float z)
{
  o.set(x, y, z);
  if (type & FRAME)
    f.makeIdentity();
  s.set(1.0f, 1.0f, 1.0f);
  type = ORIGIN;
}

//=============================================================
// mOCS::makeFrame
void mOCS::makeFrame(const mFrame& frame)
{
  o.set(0.0f, 0.0f, 0.0f);
  f.set(frame);
  s.set(1.0f, 1.0f, 1.0f);
  type = FRAME;
}

//=============================================================
// mOCS::makeScale
void mOCS::makeScale(float x, float y, float z)
{
  o.set(0.0f, 0.0f, 0.0f);
  if (type & FRAME)
    f.makeIdentity();
  s.set(x, y, z);
  type = SCALE;
}

//=============================================================
// mOCS::set
void mOCS::set(const mOCS& src)
{
  o.set(src.o);
  f.set(src.f);
  s.set(src.s);
  type = src.type;
}

//=============================================================
// mOCS::set
void mOCS::set(const mVec3& origin, 
               const mFrame& frame, 
               const mVec3& scale,
               unsigned type)
{
  (type & ORIGIN) ? o.set(origin) : o.set(0.0f, 0.0f, 0.0f);
  (type & FRAME)  ? f.set(frame)  : f.makeIdentity();
  (type & SCALE)  ? s.set(scale)  : s.set(1.0f, 1.0f, 1.0f);

  this->type = type;
}

//=============================================================
// mOCS::get
void mOCS::get(mVec3& origin, mFrame& frame, mVec3& scale) const
{
  origin.set(o);
  frame.set(f);
  scale.set(s);
}

// ============================================================
// mOCS::operator ==
bool mOCS::operator == (const mOCS& ocs) const
{
  return ( o == ocs.o &&
           s == ocs.s &&
           f == ocs.f );
}

//=============================================================
// mOCS::approxEqual
bool mOCS::approxEqual(const mOCS& ocs, float tol) const
{
  return ( o.approxEqual(ocs.o, mZeroTol) &&
           s.approxEqual(ocs.s, mZeroTol) &&
           f.approxEqual(ocs.f, mZeroTol) );
}

//=============================================================
// mOCS::inverse
void mOCS::inverse(const mOCS &ocs)
{
  if (ocs.type & ORIGIN)
    o.xformTo(mVec3(0.0f, 0.0f, 0.0f), ocs);
  else
    o.set(0.0f, 0.0f, 0.0f);

  if (ocs.type & FRAME)
    f.inverse(ocs.f);
  else if (type & FRAME)
    f.makeIdentity();

  if (ocs.type & SCALE)
    s.compDiv(mVec3(1.0f, 1.0f, 1.0f), ocs.s);
  else
    s.set(1.0f, 1.0f, 1.0f);
  
  type = ocs.type;
}

//=============================================================
// mOCS::xformTo
void mOCS::xformTo(const mOCS& ocs, const mOCS& to)
{
  mVec3 no;  
  no.xformTo(ocs.o, to);

  mFrame nf;
  if (to.type & FRAME)
    nf.xformTo(ocs.f, to.f);
  else
    nf.set(ocs.f);

  mVec3 ns;
  if (to.type & SCALE)
    ns.compDiv(ocs.s, to.s);
  else
    ns.set(ocs.s);
  
  set(no, nf, ns, ocs.type | to.type);
}

//=============================================================
// mOCS::xformFrom
void mOCS::xformFrom(const mOCS& ocs, const mOCS& from)
{
  mVec3 no;
  no.xformFrom(ocs.o, from);

  mFrame nf;
  if (from.type & FRAME)
    nf.xformFrom(ocs.f, from.f);
  else
    nf.set(ocs.f);

  mVec3 ns;
  if (from.type & SCALE)
    ns.compMult(ocs.s, from.s);  
  else
    ns.set(ocs.s);
  
  set(no, nf, ns, ocs.type | from.type);
}

//=============================================================
// mOCS::getToMatrix
void mOCS::getToMatrix(mMatrix& matrix) const
{
  float *m = matrix.m;

  m[3] = 0.0f; m[7] = 0.0f; m[11] = 0.0f; m[15] = 1.0f;

  if (type & FRAME)
  {
    m[0] = f.a[0].v[0]; m[4] = f.a[0].v[1]; m[8 ] = f.a[0].v[2];
    m[1] = f.a[1].v[0]; m[5] = f.a[1].v[1]; m[9 ] = f.a[1].v[2];
    m[2] = f.a[2].v[0]; m[6] = f.a[2].v[1]; m[10] = f.a[2].v[2];
    
    if (type & SCALE)
    {
      mVec3 ns(1.0f, 1.0f, 1.0f);
      ns.compDiv(s);
      m[0] *= ns.v[0]; m[4] *= ns.v[1]; m[8 ] *= ns.v[2]; 
      m[1] *= ns.v[0]; m[5] *= ns.v[1]; m[9 ] *= ns.v[2]; 
      m[2] *= ns.v[0]; m[6] *= ns.v[1]; m[10] *= ns.v[2]; 
    }
  }
  else
  {
    if (type & SCALE)
    {
      m[0] = 1.0f / s.v[0]; m[5] = 1.0f / s.v[1]; m[10] = 1.0f /s.v[2];
    }
    else
    {
      m[0] = m[5] = m[10] = 1.0f;
    }    
                 m[4] = 0.0f; m[8 ] = 0.0f;
    m[1] = 0.0f;              m[9 ] = 0.0f;
    m[2] = 0.0f; m[6] = 0.0f; 
  } 

  if (type & ORIGIN)
  {
    mVec3 no;
    no.xformTo(o, *this);
    m[12] = o.v[0]; m[13] = o.v[1]; m[14] = o.v[2];
  }
  else
  {
    m[12] = o.v[0]; m[13] = o.v[1]; m[14] = o.v[2];
  }
}

//=============================================================
// mOCS::getFromMatrix
void mOCS::getFromMatrix(mMatrix& matrix) const
{
  float *m = matrix.m;

  m[3] = 0.0f; m[7] = 0.0f; m[11] = 0.0f; m[15] = 1.0f;

  if (type & FRAME)
  {
    m[0] = f.a[0].v[0]; m[4] = f.a[1].v[0]; m[8 ] = f.a[2].v[0];
    m[1] = f.a[0].v[1]; m[5] = f.a[1].v[1]; m[9 ] = f.a[2].v[1];
    m[2] = f.a[0].v[2]; m[6] = f.a[1].v[2]; m[10] = f.a[2].v[2];
    
    if (type & SCALE)
    {
      m[0] *= s.v[0]; m[4] *= s.v[1]; m[8 ] *= s.v[2]; 
      m[1] *= s.v[0]; m[5] *= s.v[1]; m[9 ] *= s.v[2]; 
      m[2] *= s.v[0]; m[6] *= s.v[1]; m[10] *= s.v[2]; 
    }
  }
  else
  {
    if (type & SCALE)
    {
      m[0] = s.v[0]; m[5] = s.v[1]; m[10] = s.v[2];
    }
    else
    {
      m[0] = m[5] = m[10] = 1.0f;
    }    
                 m[4] = 0.0f; m[8 ] = 0.0f;
    m[1] = 0.0f;              m[9 ] = 0.0f;
    m[2] = 0.0f; m[6] = 0.0f; 
  } 

  m[12] = o.v[0]; m[13] = o.v[1]; m[14] = o.v[2];  
}

//****************************************************************************
//**    END MODULE mOCS.cpp
//****************************************************************************

